from datetime import date
from typing import Callable

from .period import Period
from .yearfrac import YF, YfType


class Accrual:
    """
    An accrual over a period of time with an associated function to calculate partial period accruals.

    Pass a zero argument function as the value to delay evaluation.
    Accessing the value property or comparing equality will force the evaluation of the function.

    Args:
        period: The period of time over which the accrual is calculated.
        value: The value of the accrual for the period.
        yf: A function that takes two dates and returns the fraction of a year between them.
    """

    def __init__(
        self,
        period: Period,
        value: float | Callable[[], float],
        yf: YfType,
    ):
        self.period = period
        self.yf = yf

        if isinstance(value, (float, int)):
            self._f: Callable[[], float] = lambda: value
            self._value = value
        else:
            self._f: Callable[[], float] = value
            self._value = None

    @property
    def value(self) -> float:
        value = self._value
        if value is None:
            value = self._f()
            self._value = value
        return value

    @classmethod
    def act360(cls, period: Period, value: float | Callable[[], float]) -> "Accrual":
        """Create an accrual using the actual/360 day count convention as the yf."""
        return cls(period, value, YF.actual360)

    @classmethod
    def cmonthly(cls, period: Period, value: float | Callable[[], float]) -> "Accrual":
        """Create an accrual using the calendar monthly day count convention as the yf."""
        return cls(period, value, YF.cmonthly)

    @classmethod
    def thirty360(cls, period: Period, value: float | Callable[[], float]) -> "Accrual":
        """Create an accrual using the 30/360 day count convention as the yf."""
        return cls(period, value, YF.thirty360)

    def split(self, split_date: date) -> tuple["Accrual", "Accrual"]:
        """
        Split an accrual at a given date, proportionally allocating the value.

        Args:
            split_date: The date at which to split the accrual. Must be greater than the start date and less than the end date of the accrual period.
        """
        if not (self.period.start < split_date < self.period.end):
            raise ValueError(f"Split date {split_date} must be within the accrual period {self.period}")

        total_fraction = self.yf(self.period.start, self.period.end)
        first_fraction = self.yf(self.period.start, split_date)
        second_fraction = self.yf(split_date, self.period.end)

        first_frac = first_fraction / total_fraction
        second_frac = second_fraction / total_fraction

        return (
            Accrual(Period(self.period.start, split_date), lambda: self.value * first_frac, self.yf),
            Accrual(Period(split_date, self.period.end), lambda: self.value * second_frac, self.yf),
        )

    def __add__(self, other: float | int):
        if isinstance(other, (float, int)):
            return Accrual(period=self.period, value=lambda: self.value + other, yf=self.yf)
        raise TypeError(f"Unsupported operand type(s) for +: 'Accrual' and '{type(other).__name__}'")

    def __radd__(self, other: float | int):
        return self.__add__(other)

    def __sub__(self, other: float | int):
        if isinstance(other, (float, int)):
            return Accrual(period=self.period, value=lambda: self.value - other, yf=self.yf)
        raise TypeError(f"Unsupported operand type(s) for -: 'Accrual' and '{type(other).__name__}'")

    def __mul__(self, other: float | int):
        if isinstance(other, (float, int)):
            return Accrual(period=self.period, value=lambda: self.value * other, yf=self.yf)
        raise TypeError(f"Unsupported operand type(s) for *: 'Accrual' and '{type(other).__name__}'")

    def __rmul__(self, other: float | int):
        return self.__mul__(other)

    def __truediv__(self, other: float | int):
        if isinstance(other, (float, int)):
            return Accrual(period=self.period, value=lambda: self.value / other, yf=self.yf)
        raise TypeError(f"Unsupported operand type(s) for /: 'Accrual' and '{type(other).__name__}'")

    def __neg__(self):
        return Accrual(period=self.period, value=lambda: -self.value, yf=self.yf)
    
    def __repr__(self) -> str:
        return f"Accrual(period={self.period.__repr__()}, value={self._value if self._value else '() -> float'}, yf={self.yf.__repr__()})"
    
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Accrual):
            return NotImplemented
        return (
            self.period == value.period
            and self.value == value.value
            and self.yf == value.yf
        )



import datetime
from abc import ABC, abstractmethod
from dataclasses import dataclass
from itertools import takewhile
from typing import Callable, Iterable, Iterator

from orcaset import Node, cached_generator


class Payment:
    """
    Represents a payment made on a specific date.

    Pass a zero argument function as the value to lazily delay evaluation.
    Accessing the value property or comparing equality will force the evaluation of the function.
    """

    def __init__(self, date: datetime.date, value: float | Callable[[], float]):
        self.date = date
        if isinstance(value, (float, int)):
            self._f = lambda: value
            self._value = value
        else:
            self._f = value
            self._value = None

    @property
    def value(self) -> float:
        value = self._value
        if value is None:
            value = self._f()
            self._value = value
        return value

    def __add__(self, other: float | int):
        if isinstance(other, (float, int)):
            return Payment(date=self.date, value=lambda: self.value + other)
        raise TypeError(f"Cannot add {type(other)} to {type(self)}. Only float or int is allowed.")

    def __radd__(self, other: float | int):
        return self.__add__(other)

    def __sub__(self, other: float | int):
        if isinstance(other, (float, int)):
            return Payment(date=self.date, value=lambda: self.value - other)
        raise TypeError(f"Cannot subtract {type(other)} from {type(self)}. Only float or int is allowed.")

    def __mul__(self, other: float | int):
        if isinstance(other, (float, int)):
            return Payment(date=self.date, value=lambda: self.value * other)
        raise TypeError(f"Cannot multiply {type(other)} with {type(self)}. Only float or int is allowed.")

    def __rmul__(self, other: float | int):
        return self.__mul__(other)

    def __truediv__(self, other: float | int):
        if isinstance(other, (float, int)):
            return Payment(date=self.date, value=lambda: self.value / other)
        raise TypeError(f"Cannot divide {type(self)} by {type(other)}. Only float or int is allowed.")

    def __neg__(self):
        return Payment(date=self.date, value=lambda: -self.value)
    
    def __repr__(self) -> str:
        return f"Payment(date={self.date}, value={self._value.__repr__() if self._value is not None else '<unevaluated>'})"
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Payment):
            return NotImplemented
        return self.date == other.date and self.value == other.value


def _combine_payment_series(
    first: Iterable[Payment], second: Iterable[Payment], operator: Callable[[float, float], float]
) -> Iterable[Payment]:
    """
    Combine two payment series using the provided operator.
    Payments on the same date are combined using the operator.
    """
    first = iter(first)
    second = iter(second)

    first_pmt = next(first, None)
    second_pmt = next(second, None)
    while first_pmt is not None or second_pmt is not None:
        if first_pmt is None:
            yield second_pmt  # type: ignore
            second_pmt = next(second, None)
        elif second_pmt is None:
            yield first_pmt
            first_pmt = next(first, None)
        elif first_pmt.date < second_pmt.date:
            yield first_pmt
            first_pmt = next(first, None)
        elif first_pmt.date > second_pmt.date:
            yield second_pmt
            second_pmt = next(second, None)
        else:
            yield Payment(date=first_pmt.date, value=lambda fp=first_pmt, sp=second_pmt: operator(fp.value, sp.value))
            first_pmt = next(first, None)
            second_pmt = next(second, None)


@dataclass
class PaymentSeriesBase[P](Node[P], ABC):
    """A series of `Payment` objects. Subclasses must override `_payments` to provide consecutive values by ascending date."""

    @abstractmethod
    def _payments(self) -> Iterable[Payment]: ...

    @cached_generator
    def __iter__(self) -> Iterator[Payment]:
        yield from self._payments()

    def on(self, dt: datetime.date) -> float:
        """
        Get the payment at a given date. Returns zero if no payment on the given date.
        """
        pmt = next((pmt for pmt in self if pmt.date == dt), None)
        if pmt is None:
            return 0
        else:
            return pmt.value

    def over(self, from_date: datetime.date, to_date: datetime.date) -> float:
        """
        Get the total payment from and excluding `from_date` to and including `to_date`.
        Returns zero if no payments are made in the period.
        """
        total = 0
        for pmt in takewhile(lambda pmt: pmt.date <= to_date, self):
            if from_date < pmt.date <= to_date:
                total += pmt.value
        return total

    def after(self, dt: datetime.date) -> "PaymentSeries":
        """Get a new `PaymentSeries` containing payments after the given date."""
        return PaymentSeries(payment_series=(pmt for pmt in self if pmt.date > dt))

    def __add__(self, other: "PaymentSeriesBase | float | int") -> "PaymentSeries":
        if isinstance(other, PaymentSeriesBase):
            return PaymentSeries(payment_series=_combine_payment_series(self, other, lambda a, b: a + b))
        elif isinstance(other, (float, int)):
            return PaymentSeries(payment_series=(pmt + other for pmt in self))
        else:
            raise TypeError(f"Cannot add {type(other)} to {type(self)}. Only PaymentSeriesBase, float or int is allowed.")

    def __radd__(self, other: "PaymentSeriesBase | float | int"):
        return self.__add__(other)

    def __sub__(self, other: "PaymentSeriesBase | float | int") -> "PaymentSeries":
        if isinstance(other, PaymentSeriesBase):
            return PaymentSeries(payment_series=_combine_payment_series(self, other, lambda a, b: a - b))
        elif isinstance(other, (float, int)):
            return PaymentSeries(payment_series=(pmt - other for pmt in self))
        else:
            raise TypeError(f"Cannot subtract {type(other)} from {type(self)}. Only PaymentSeriesBase, float or int is allowed.")

    def __mul__(self, other: "PaymentSeriesBase | float | int") -> "PaymentSeries":
        if isinstance(other, PaymentSeriesBase):
            return PaymentSeries(payment_series=_combine_payment_series(self, other, lambda a, b: a * b))
        elif isinstance(other, (float, int)):
            return PaymentSeries(payment_series=(pmt * other for pmt in self))
        else:
            raise TypeError(f"Cannot multiply {type(other)} with {type(self)}. Only PaymentSeriesBase, float or int is allowed.")

    def __rmul__(self, other: "PaymentSeriesBase | float | int"):
        return self.__mul__(other)

    def __truediv__(self, other: "PaymentSeriesBase | float | int") -> "PaymentSeries":
        if isinstance(other, PaymentSeriesBase):
            return PaymentSeries(payment_series=_combine_payment_series(self, other, lambda a, b: a / b))
        elif isinstance(other, (float, int)):
            return PaymentSeries(payment_series=(pmt / other for pmt in self))
        else:
            raise TypeError(f"Cannot divide {type(self)} by {type(other)}. Only PaymentSeriesBase, float or int is allowed.")

    def __neg__(self) -> "PaymentSeries":
        """Return a new PaymentSeries that negates the payments of `self`"""
        return PaymentSeries(payment_series=(-pmt for pmt in self))


@dataclass
class PaymentSeries[P](PaymentSeriesBase[P]):
    """A series of payments that takes a `payment_series: Iterable[Pmt]` constructor variable."""

    payment_series: Iterable[Payment]

    def _payments(self) -> Iterable[Payment]:
        yield from self.payment_series



from datetime import date
import pytest
from orcaset.financial import Accrual, AccrualSeries, Period, YF


class TestAccrueNoOverlap:
    """Test accrue method when query period doesn't overlap with accrual series."""
    
    def test_query_before_all_accruals(self):
        """Query period entirely before all accruals."""
        accruals = [
            Accrual.act360(Period(date(2023, 2, 1), date(2023, 2, 15)), 100.0),
            Accrual.act360(Period(date(2023, 2, 15), date(2023, 3, 1)), 200.0),
        ]
        series = AccrualSeries(series=accruals)

        result = series.accrue(date(2023, 1, 1), date(2023, 2, 1))
        assert result == pytest.approx(0.0)
    
    def test_query_after_all_accruals(self):
        """Query period entirely after all accruals."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 15)), 100.0),
            Accrual.act360(Period(date(2023, 1, 15), date(2023, 2, 1)), 200.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 2, 2), date(2023, 3, 31))
        assert result == pytest.approx(0.0)
    
    def test_query_in_gap_between_accruals(self):
        """Query period in gap between non-consecutive accruals."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 15)), 100.0),
            Accrual.act360(Period(date(2023, 3, 1), date(2023, 3, 15)), 200.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 16), date(2023, 3, 1))
        assert result == pytest.approx(0.0)


class TestAccruePartialOverlap:
    """Test accrue method when query period partially overlaps with accruals."""
    
    def test_query_starts_before_ends_during_first_accrual(self):
        """Query starts before series, ends during first accrual."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 10), date(2023, 1, 20)), 100.0),
            Accrual.act360(Period(date(2023, 1, 20), date(2023, 1, 30)), 200.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 5), date(2023, 1, 15))
        
        expected = 100.0 * 5 / 10
        assert result == pytest.approx(expected)
    
    def test_query_starts_during_ends_after_last_accrual(self):
        """Query starts during accrual, ends after series."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 10)), 100.0),
            Accrual.act360(Period(date(2023, 1, 10), date(2023, 1, 20)), 200.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 15), date(2023, 1, 30))
        
        expected = 200.0 * 5 / 10
        assert result == pytest.approx(expected)
    
    def test_query_within_single_accrual(self):
        """Query period completely within a single accrual."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 31)), 310.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 10), date(2023, 1, 20))
        
        expected = 310.0 * 10 / 30
        assert result == pytest.approx(expected)


class TestAccrueCompleteOverlap:
    """Test accrue method when query period completely overlaps with accruals."""
    
    def test_query_exactly_matches_single_accrual(self):
        """Query period exactly matches a single accrual period."""
        accruals = [Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 31)), 310.0)]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 1), date(2023, 1, 31))
        assert result == pytest.approx(310.0)
    
    def test_query_encompasses_entire_series(self):
        """Query period encompasses entire accrual series."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 10), date(2023, 1, 20)), 100.0),
            Accrual.act360(Period(date(2023, 1, 20), date(2023, 1, 30)), 200.0),
            Accrual.act360(Period(date(2023, 1, 30), date(2023, 2, 10)), 300.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 5), date(2023, 2, 15))
        assert result == pytest.approx(600.0)
    
    def test_single_accrual_encompasses_query(self):
        """Single accrual period encompasses entire query period."""
        accruals = [Accrual.act360(Period(date(2023, 1, 1), date(2023, 3, 1)), 600.0)]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 15), date(2023, 2, 15))
        assert result == pytest.approx(600.0 * 31 / 59)


class TestAccrueMultiAccrualScenarios:
    """Test accrue method with complex multi-accrual scenarios."""
    
    def test_query_spans_accruals_with_gaps(self):
        """Query period spans accruals with gaps between them."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 10)), 100.0),
            Accrual.act360(Period(date(2023, 1, 20), date(2023, 1, 30)), 200.0),
            Accrual.act360(Period(date(2023, 2, 10), date(2023, 2, 20)), 300.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 1), date(2023, 2, 20))
        assert result == pytest.approx(600.0)
    
    def test_mixed_partial_complete_overlaps(self):
        """Query with mixed partial and complete overlaps across multiple accruals."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 11)), 100.0),   # 10 days
            Accrual.act360(Period(date(2023, 1, 11), date(2023, 1, 21)), 200.0),  # 10 days - complete
            Accrual.act360(Period(date(2023, 1, 21), date(2023, 2, 1)), 330.0),   # 11 days - partial
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 6), date(2023, 1, 26))
        assert result == pytest.approx((100.0 * 5/10) + 200.0 + (330.0 * 5/11))

    def test_query_touches_accrual_boundaries(self):
        """Query period boundaries align with accrual boundaries."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 15)), 150.0),
            Accrual.act360(Period(date(2023, 1, 15), date(2023, 2, 1)), 170.0),
            Accrual.act360(Period(date(2023, 2, 1), date(2023, 2, 15)), 150.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 15), date(2023, 2, 1))
        assert result == pytest.approx(170.0)


class TestAccrueEdgeCases:
    """Test accrue method edge cases."""
    
    def test_empty_accrual_series(self):
        """Test accrue with empty accrual series."""
        series = AccrualSeries(series=[])
        
        result = series.accrue(date(2023, 1, 1), date(2023, 1, 31))
        assert result == pytest.approx(0.0)
    
    def test_query_dates_equal(self):
        """Test accrue when query start and end dates are equal (zero period)."""
        accruals = [Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 31)), 310.0)]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 15), date(2023, 1, 15))
        assert result == pytest.approx(0.0)
    
    def test_negative_accrual_values(self):
        """Test accrue with negative accrual values."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 15)), -100.0),
            Accrual.act360(Period(date(2023, 1, 15), date(2023, 2, 1)), 200.0),
            Accrual.act360(Period(date(2023, 2, 1), date(2023, 2, 15)), -50.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 1), date(2023, 2, 15))
        assert result == pytest.approx(50.0)
    
    def test_different_year_fraction_methods(self):
        """Test accrue with accruals using different year fraction methods."""
        accruals = [
            Accrual(Period(date(2022, 12, 31), date(2023, 1, 31)), 150.0, YF.actual360),
            Accrual(Period(date(2023, 1, 31), date(2023, 2, 28)), 160.0, YF.thirty360),
            Accrual(Period(date(2023, 2, 28), date(2023, 3, 31)), 150.0, lambda _, __: 1),  # Constant
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 10), date(2023, 3, 10))
        
        first_partial = 150.0 * YF.actual360(date(2023, 1, 10), date(2023, 1, 31))/ YF.actual360(date(2022, 12, 31), date(2023, 1, 31))
        second_complete = 160.0
        third_partial = 150.0
        
        expected = first_partial + second_complete + third_partial
        assert result == pytest.approx(expected)
    
    def test_single_day_accruals(self):
        """Test accrue with single-day accrual periods."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 2)), 100.0),
            Accrual.act360(Period(date(2023, 1, 2), date(2023, 1, 3)), 200.0),
            Accrual.act360(Period(date(2023, 1, 3), date(2023, 1, 4)), 300.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 1), date(2023, 1, 4))
        assert result == pytest.approx(600.0)
    
    def test_query_exactly_at_accrual_boundary(self):
        """Test accrue when query boundaries exactly match accrual boundaries."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 15)), 150.0),
            Accrual.act360(Period(date(2023, 1, 15), date(2023, 2, 1)), 170.0),
            Accrual.act360(Period(date(2023, 2, 1), date(2023, 2, 15)), 150.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 15), date(2023, 2, 1))
        assert result == pytest.approx(170.0)


class TestAccrueComplexScenarios:
    """Test accrue method with complex real-world scenarios."""
    
    def test_irregular_accrual_periods(self):
        """Test with irregular, non-uniform accrual periods."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 5)), 40.0),     # 4 days
            Accrual.act360(Period(date(2023, 1, 5), date(2023, 1, 25)), 400.0),   # 20 days
            Accrual.act360(Period(date(2023, 1, 25), date(2023, 1, 27)), 6.0),    # 2 days
            Accrual.act360(Period(date(2023, 1, 27), date(2023, 2, 15)), 380.0),  # 19 days
        ]
        series = AccrualSeries(series=accruals)
        
        # Query overlaps all periods partially and completely
        result = series.accrue(date(2023, 1, 3), date(2023, 2, 10))
        
        # Expected: partial first + complete second + complete third + partial fourth
        first_partial = 40.0 * 2/4        # Jan 3-5: 2 out of 4 days
        second_complete = 400.0           # Complete
        third_complete = 6.0              # Complete
        fourth_partial = 380.0 * 14/19    # Jan 27-Feb 10: 14 out of 19 days
        
        expected = first_partial + second_complete + third_complete + fourth_partial
        assert result == pytest.approx(expected)
    
    def test_many_small_accruals(self):
        """Test with many small consecutive accruals."""
        accruals = []
        total_expected = 0.0
        
        # Create 30 daily accruals with incrementing values
        for i in range(30):
            start_date = date(2023, 1, 1 + i)
            end_date = date(2023, 1, 2 + i)
            value = float(i + 1) * 10  # 10, 20, 30, ..., 300
            accruals.append(Accrual.act360(Period(start_date, end_date), value))
            total_expected += value
        
        series = AccrualSeries(series=accruals)
        
        # Query encompasses all accruals
        result = series.accrue(date(2023, 1, 1), date(2023, 1, 31))
        assert result == pytest.approx(total_expected)
    
    def test_overlapping_query_boundaries_complex(self):
        """Test complex scenario with query boundaries creating multiple splits."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 20)), 200.0),   # 19 days
            Accrual.act360(Period(date(2023, 1, 20), date(2023, 2, 10)), 210.0),  # 21 days
            Accrual.act360(Period(date(2023, 2, 10), date(2023, 3, 1)), 190.0),   # 19 days
        ]
        series = AccrualSeries(series=accruals)
        
        # Query starts mid-first, ends mid-last
        result = series.accrue(date(2023, 1, 10), date(2023, 2, 20))
        
        # Expected: partial first + complete second + partial third
        first_partial = 200.0 * 10/19     # Jan 10-20: 10 out of 19 days
        second_complete = 210.0           # Complete
        third_partial = 190.0 * 10/19     # Feb 10-20: 10 out of 19 days
        
        expected = first_partial + second_complete + third_partial
        assert result == pytest.approx(expected)


class TestAccrueLazyEvaluation:
    """Test that accrue method respects lazy evaluation of accrual values."""
    
    def test_lazy_evaluation_only_evaluates_overlapping_accruals(self):
        """Test that only accruals overlapping with query period are evaluated."""
        from unittest.mock import Mock
        
        # Create mocks for values
        before_mock = Mock(return_value=100.0)
        overlap_mock = Mock(return_value=200.0)
        after_mock = Mock(return_value=300.0)
        
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 10)), before_mock),   # Before query
            Accrual.act360(Period(date(2023, 1, 15), date(2023, 1, 25)), overlap_mock), # Overlaps query
            Accrual.act360(Period(date(2023, 2, 1), date(2023, 2, 10)), after_mock),   # After query
        ]
        series = AccrualSeries(series=accruals)
        
        # Query only overlaps middle accrual
        result = series.accrue(date(2023, 1, 12), date(2023, 1, 30))
        
        # Only the overlapping accrual should be evaluated
        before_mock.assert_not_called()
        overlap_mock.assert_called_once()
        after_mock.assert_not_called()
        
        # Should get partial value from middle accrual
        expected = 200.0 * 10/10  # Jan 15-25: all 10 days within query period
        assert result == pytest.approx(expected)
    
    def test_lazy_evaluation_with_partial_overlaps(self):
        """Test lazy evaluation with accruals that need splitting."""
        from unittest.mock import Mock
        
        value_mock = Mock(return_value=300.0)
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 31)), value_mock),
        ]
        series = AccrualSeries(series=accruals)
        
        # Query partial overlap that requires splitting
        result = series.accrue(date(2023, 1, 10), date(2023, 1, 20))
        
        # Value should be evaluated when split occurs
        value_mock.assert_called_once()
        
        # Should get partial value (10 out of 30 days)
        expected = 300.0 * 10/30
        assert result == pytest.approx(expected)


class TestAccrueInvalidInputs:
    """Test accrue method with invalid inputs."""
    
    def test_dt1_after_dt2(self):
        """Test accrue when start date is after end date."""
        accruals = [
            Accrual.act360(Period(date(2023, 1, 1), date(2023, 1, 31)), 310.0),
        ]
        series = AccrualSeries(series=accruals)
        
        result = series.accrue(date(2023, 1, 20), date(2023, 1, 10))
        assert result == pytest.approx(310.0 * 10 / 30)